from django.template import Context
from django.template.base import Variable
from django.test import TestCase

from .template.choices import AttributeChoices
from .template.attributes import Attribute
from .template.builtins import register
from .template.components import Slot


class ChoiceTestCase(TestCase):

    def test_format(self):
        class FooChoices(AttributeChoices):
            bar = 'bar'

        self.assertEqual(format(FooChoices.bar), 'bar')

    def test_format_repr(self):
        class FooChoices(AttributeChoices):
            bar = 1

        self.assertIsInstance(format(FooChoices.bar), str)


class AttributeTestCase(TestCase):

    def setUp(self):
        class CustomChoices(AttributeChoices):
            foo = 'bar'

        self.choices = CustomChoices

    def test_is_instance(self):
        self.assertIsInstance(Attribute(), Attribute)

    def test_choices(self):
        attr = Attribute(choices=self.choices)
        x = Variable('attr')
        c = Context({'attr': 'foo'})
        self.assertEqual(attr.resolve(x, c), 'bar')

    def test_choice_does_not_exist(self):
        attr = Attribute(choices=self.choices, required=True)
        x = Variable('attr')
        c = Context({'attr': 'ERROR'})
        self.assertRaises(Attribute.ChoiceDoesNotExist, lambda: attr.resolve(x, c))

        attr = Attribute(choices=self.choices, required=True)
        x = Variable('attr')
        c = Context({'attr': 'ERROR'})
        self.assertRaises(Attribute.ChoiceDoesNotExist, lambda: attr.get_choice('x', raise_exception=True))

    def test_choices_default(self):
        attr = Attribute(choices=self.choices, default=self.choices.foo)

        try:
            # Not placing the attr inside context on purpose
            x = Variable('attr')
            c = Context()
            value = attr.resolve(x, c)
        except Attribute.VariableDoesNotExist:
            value = attr.default

        self.assertEqual(value, 'bar')

    def test_variable_does_not_exist(self):
        attr = Attribute(choices=self.choices)

        def test_func():
            x = Variable('attr')
            c = Context()
            return attr.resolve(x, c)

        self.assertRaises(Attribute.VariableDoesNotExist, test_func)

    def test_choices_are_required_and_value_is_none(self):
        attr = Attribute(choices=self.choices, required=True)

        def test_func():
            try:
                x = Variable('attr')
                c = Context({'attr': None})
                return attr.resolve(x, c)
            except Attribute.VariableDoesNotExist:
                return attr.default

        self.assertRaises(Attribute.RequiredValue, test_func)

    def test_choices_are_not_enum_type(self):
        class Choices:
            bar = 1

        def test_with_class():
            return Attribute(choices=Choices)

        def test_with_list():
            return Attribute(choices=[{'foo': 'bar'}])

        self.assertRaises(Attribute.RequiredValue, test_with_class)
        self.assertRaises(Attribute.RequiredValue, test_with_list)

    def test_set_context_name(self):
        attr = Attribute(choices=self.choices)
        attr.set_context_name('foo')
        self.assertEqual(attr.context_name, 'foo')


class BuiltinsTestCase(TestCase):

    def setUp(self) -> None:
        self.BUILTIN_TAGS = [
            ('slot', Slot)
        ]

    def test_default_tags(self):
        for name, i in self.BUILTIN_TAGS:
            self.assertIn(name, register.tags)
            self.assertTrue(issubclass(register.tags[name].__wrapped__, Slot))


class ComponentTestCase(TestCase):
    pass


class ContextTestCase(TestCase):
    pass


class HelperTestCase(TestCase):
    pass


class LibraryTestCase(TestCase):
    pass


class NodeTestCase(TestCase):
    pass


class ParserTestCase(TestCase):
    pass


class WrapperTestCase(TestCase):
    pass



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class ComponentTagsConfig(AppConfig):
    name = 'component_tags'
    verbose_name = _('Django Component Tags')



from django.template.library import Library as BaseLibrary

from .wrappers import component_wrapper
from .nodes import BaseComponent

__all__ = ['Library']


class Library(BaseLibrary):
    """
    A custom class for registering template components, tags and filters.
    The filter, simple_tag, and inclusion_tag methods provide a convenient
    way to register callables/components as tags.

    Examples
    --------
    There is a lot of ways to register a component tag:

    .. code-block:: python

        register.tag('name', ComponentClass)

        @register.tag
        class ComponentClass:
            pass

        @register.tag()
        class ComponentClass:
            pass

        @register.tag('name')
        class ComponentClass:
            pass

        register.tag_function(ComponentClass)

        register.component('name', ComponentClass)

        @register.component
        class ComponentClass:
            pass

    Notes
    ----
    - Pycharm doesnt recognize @register.component, therefore the recommended way to register all components
    using Pycharm is with `register.tag('name', ComponentClass)`
    - Does not interfere with the current Django Library behavior, but it decorates the callable function
    with the component wrapper if this is a component tag.
    """

    def tag_function(self, func):
        if isinstance(func, BaseComponent):
            name = getattr(func, "_decorated_function", func).__name__.lower()
            name, func = component_wrapper(name, func)
            self.tags[name] = func
            return func
        return super().tag_function(func)

    def tag(self, name=None, compile_function=None):
        # @register.tag()
        if name is None and compile_function is None:
            return self.tag_function
        if name is not None and compile_function is None:
            # @register.tag
            if callable(name):
                if isinstance(name, BaseComponent):
                    component_name = name.__name__.lower()
                    name, func = component_wrapper(component_name, name)
                    self.tags[name] = func
                    return func
                return super().tag_function(name)
            # @register.tag('foobar') or @register.tag(name='foobar')
            else:
                tag_func = super().tag

                def dec(f):
                    if isinstance(f, BaseComponent):
                        return tag_func(*component_wrapper(name, f))
                    return tag_func(name, f)

                return dec
        # register.tag('foobar', foobar)
        elif name is not None and compile_function is not None:
            if isinstance(compile_function, BaseComponent):
                name, func = component_wrapper(name, compile_function)
                self.tags[name] = func
                return func
            self.tags[name] = compile_function
            return compile_function
        else:
            raise ValueError("Unsupported arguments to Library.tag: (%r, %r)" % (name, compile_function))

    def component(self, name=None, compile_function=None):
        return self.tag(name=name, compile_function=compile_function)



from .attributes import Attribute
from .choices import AttributeChoices
from .context import TagContext, ComponentContext
from .library import Library
from .nodes import ComponentNode


__all__ = [
    'Library',
    'Attribute',
    'AttributeChoices',
    'Component',
    'ComponentContext',
    'TagContext',
]


Component = ComponentNode



from typing import Optional, Union

from django.template import Context, RequestContext
from django.utils.safestring import SafeText, SafeString

__all__ = [
    'format_value',
    'format_classes',
    'format_attributes',
]


def format_value(value, context=None):
    """
    Resole the Variable/FilterExpression value else nothing happens
    """
    try:
        return value.resolve(context)
    except AttributeError:
        return value


def format_classes(classes: Optional[list], context=Union[Context, RequestContext]) -> Optional[SafeText]:
    """
    Join all classes as a string
    """
    # Is format_classes already called before?
    if isinstance(classes, SafeString):
        return classes

    # Do not print anything if there are no html classes
    if not classes:
        return None

    elements = []
    for value in classes:
        value = format_value(value, context)
        if value is not None:
            elements.append(value)

    return SafeText(" ".join(elements))


# TODO: remove this function
# def format_options(options, context=None) -> SafeText:
#     elements = []
#     for (key, value) in options.items():
#         value = format_value(value, context)
#         if value is not None:
#             elements.append("{}: {}".format(key, value))
#     return SafeText("; ".join(elements).lower())


def format_attributes(properties, context=None) -> SafeText:
    """
    Join all attributes as a string
    """
    # Is format_attributes already called before?
    if isinstance(properties, SafeString):
        return properties

    elements = []
    for (key, value) in properties.items():
        # Is this a class attribute
        if key == 'class':
            value = format_classes(value, context)
        else:
            value = format_value(value, context)
        if value is not None:
            elements.append(f'{key}="{value}"')

    return SafeText(" ".join(elements))



from inspect import getmembers
from copy import copy

from django.forms.widgets import Media
from django.template.base import Node, NodeList

from .media import add_media
from .attributes import Attribute
from .context import ComponentContext


__all__ = ['ComponentNode', 'BaseComponent', 'Meta', 'Media']


class TemplateIsNull(Exception):
    pass


class Meta(Media):
    """
    Meta definition of component tags, currently only used to declare template information
    """
    def __init__(self, meta=None, css=None, js=None):
        super().__init__(meta, css, js)
        self.template_name = getattr(meta, 'template_name', None)


def meta_property(cls):
    """
    Get the media property of the superclass, if it exists
    """
    def _meta(self):

        sup_cls = super(cls, self)
        try:
            # noinspection PyUnresolvedReferences
            base = sup_cls.meta
        except AttributeError:
            base = Meta()

        # Get the media definition for this class
        definition = getattr(cls, 'Meta', None)

        if definition:
            extend = getattr(definition, 'extend', True)
            if extend:
                if extend is True:
                    m = base
                else:
                    m = Meta()
                    for medium in extend:
                        m = m + base[medium]
                extended = m + Meta(definition)
            else:
                extended = Meta(definition)

            template_name = getattr(definition, 'template_name', None)
            if template_name:
                setattr(extended, 'template_name', template_name)
            return extended
        return base
    return property(_meta)


class BaseComponent(type):
    """Metaclass for all component nodes."""

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)

        if 'meta' not in attrs:
            new_class.meta = meta_property(new_class)

        return new_class


class ComponentNode(Node, metaclass=BaseComponent):
    """
    Components are used to mark up the start of an HTML element
    and they are usually enclosed in angle brackets.
    """

    TemplateIsNull = TemplateIsNull

    def __init__(self, tag_name: str, nodelist: NodeList, options: dict, slots: dict, *args,
                 isolated_context: bool = True, **kwargs):
        self.tag_name = tag_name
        self.nodelist = nodelist
        self.attrs = kwargs
        self.slots = slots
        self.options = options
        self.isolated_context = isolated_context

    def get_template_name(self):
        return getattr(self.meta, 'template_name', None)

    def get_template(self, context):
        template_name = self.get_template_name()

        if not template_name:
            raise self.TemplateIsNull(f'[{self.tag_name.title()}] component does not have a template assigned.')

        return context.template.engine.get_template(template_name)

    def get_context_data(self, context):
        return ComponentContext(self.nodelist, initial=context, isolated=self.isolated_context)

    def render(self, context):
        add_media(context, self.meta)
        template = self.get_template(context)

        # Does this quack like a Template?
        if not callable(getattr(template, 'render', None)):
            # If not, try the cache and select_template().
            template_name = template or ()
            if isinstance(template_name, str):
                template_name = (template_name,)
            else:
                template_name = tuple(template_name)
            cache = context.render_context.dicts[0].setdefault(self, {})
            template = cache.get(template_name)

            if template is None:
                template = context.template.engine.select_template(template_name)
                cache[template_name] = template

        # Use the base.Template of a backends.django.Template.
        elif hasattr(template, 'template'):
            template = template.template

        attrs = self.attrs.copy()

        # Do not use original context since we are updating values inside this function
        _context = self.get_context_data(copy(context))

        # Class attributes
        class_attrs = getmembers(self, lambda a: isinstance(a, Attribute))

        while class_attrs:
            key, attr = class_attrs.pop()

            if not attr.context_name:
                attr.set_context_name(key)

            try:
                value = attr.resolve(attrs.pop(key), context)
            except KeyError:
                value = attr.default

            key = attr.context_name

            if attr.as_context:
                _context[key] = value
            elif attr.as_class:
                _context.add_class(value)
            else:
                _context.add_attribute(key, value)

        # Attribute variables
        for name, value in attrs.items():
            _context.add_attribute(name, value)

        # Option variables
        for name, value in self.options.items():
            _context[name] = value.resolve(context)

        context = _context.make()  # Slots should only have access to parent context

        # Slot nodes
        for name, value in self.slots.items():
            _context[name] = value.render(context)

        return template.render(context)



from .parser import parse_component
from functools import wraps

__all__ = ['component_wrapper']


def component_wrapper(name, component_node):
    """
    Component wrapper to parse all component args
    """
    @wraps(component_node)
    def func(parser, token):
        nodelist = parser.parse(('end%s' % name,))
        parser.delete_first_token()
        tag_name, args, kwargs, options, slots, isolated_context = parse_component(nodelist, token, parser)
        return component_node(tag_name, nodelist, options, slots, *args, isolated_context=isolated_context, **kwargs)
    return name, func



from django.template.base import Variable, NodeList

from .nodes import ComponentNode

__all__ = ['Slot']


class Slot(ComponentNode):
    """
    Can be used inside any other component to create dynamic blocks inside components.

    Attributes
    ----------
    name: str
        name used to identify the slot inside the component tag

    Examples
    --------

        .. code-block::

            # tags/button.html
            <span>{{ slot_foo }}</span>
            <button {{ attributes }}>
                {{ nodelist }}
            </button>

            # base.html
            {% button type="button" %}
                {% slot 'foo' %}Slot 1{% endslot %}
                Button 1
            {% endbutton %}

            # Output:
            <span>Slot 1</span>
            <button type="button"></button>
    """

    def __init__(self, *args, **kwargs):
        tag_name, nodelist, options, slots, name = args
        kwargs['isolated_context'] = True  # Make sure slot has the parent context
        super().__init__(tag_name, nodelist, options, slots, **kwargs)

        # TODO:
        #  there should be a better way to get the variable value
        #  but name should be declared inside this constructor
        self.slot_name = getattr(name, 'var', None)

    class Meta:
        template_name = 'component_tags/slot.html'



from enum import Enum
from typing import Optional, Union

from django.template.base import Variable, FilterExpression, VariableDoesNotExist

__all__ = ['Attribute']


class ChoiceDoesNotExist(Exception):
    """
    Raised when the specified choice is not a choice's member
    """
    pass


class RequiredValue(Exception):
    """
    Raised when a variable value is required
    """
    pass


class Attribute:
    """
    Template Components have attributes; these are additional values that configure the elements or adjust their
    behavior in various ways to meet the criteria the users want.

    Attributes
    ----------
    name : str
        name used to identify it inside the context
    choices: AttributeChoices
        choices are used to specify the attribute correct value from a list of available options
    default: Any
        default attribute value
    required: bool
        set if the attribute value is required
    as_class: bool
        set the attribute as "class type" inside the template
    as_context: bool
        set the attribute as "context type" inside the template

    Examples
    --------
    **Attribute as class**

    Considering the following definition:

    .. code-block:: python

        # templatetags/custom_tags.py

        @register.component
        class Button(template.Component):
            class ColorChoices(template.AttributeChoices):
                primary = 'btn-primary'
                secondary = 'btn-secondary'

            color = template.Attribute(choices=ColorChoices, default=ColorChoices.primary, as_class=True)

            class Meta:
                template_name = 'components/button.html'

    And the following template

    .. code-block:: 

        # templates/components/button.html

        <button {{ attributes }}></button>

    Output

    .. code-block::

        # Setting a value
        {% button color="secondary" %}{% endbutton %}

        # output:
        <button class="btn-secondary"></button>

        # Without setting a value
        {% button %}{% endbutton %}

        # output:
        <button class="btn-primary"></button>

    **Attribute as context**

    Considering the following definition:

    .. code-block:: python

        class Button(template.Component):
            href = template.Attribute(default="#", as_context=True)

    You can use this attribute inside the component context as follows:

    .. code-block::

        # Setting a value
        {% button href="/some-url" %}
            {{ href }} # output: "/some-url"
        {% endbutton %}

        # Without setting a value
        {% button %}
            {{ href }} # output: "#"
        {% endbutton %}
    """

    ChoiceDoesNotExist = ChoiceDoesNotExist
    VariableDoesNotExist = VariableDoesNotExist
    RequiredValue = RequiredValue

    def __init__(self, choices=None, default=None, required: bool = False, context_name: str = None,
                 as_class: bool = False, as_context: bool = False):
        self.context_name = context_name

        try:
            if choices and not issubclass(choices, Enum):
                raise RequiredValue("choices must be Enum's subclass")
        except TypeError:
            raise RequiredValue("choices must be Enum's subclass")

        self.choices = choices
        self.required = required
        self.default = default
        self.as_class = as_class
        self.as_context = as_context

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, value: Optional[Enum]):
        """
        Sets the default value and checks if can be applied.
        """
        self._default = self.check_value(value) if value is not None else value

    def get_member_choices(self):
        """
        Get the member choices as list
        """
        return [key for key, value in self.choices.__members__.items()]

    def get_choice(self, key: Union[Enum, str], raise_exception: bool = False):
        """
        Get the member choice value using the Enum format
        """
        if not raise_exception:
            raise_exception = self.required

        try:
            # noinspection PyCallingNonCallable,PyArgumentList
            return format(key if isinstance(key, self.choices) else self.choices[key])
        except (KeyError, ValueError):
            if raise_exception:
                raise ChoiceDoesNotExist(f'{key} is not an available choice, choices are: {self.get_member_choices()}')

    def check_value(self, value, raise_exception: bool = False):
        """
        Verify if the current value is valid based on different criteria
        """
        if self.required and value is None:
            raise RequiredValue(f'{self.context_name} should be different than null')
        return self.get_choice(value, raise_exception) if self.choices else value

    def set_context_name(self, value: str):
        """
        The name of the attribute used as context name.

        Sometimes declaring the names it is not enough to get the result needed, specially when you want to
        work with javascript.
        
        Examples
        --------
        
        **As a class instance:**

        .. code-block:: python
            
            class Button(template.Component):
                href = template.Attribute(default="#", as_context=True)

        Output:

        .. code-block::

                # base.html

                # Setting a value
                {% button href="foo-bar/" %}{% button %}

                # Output:
                <button href="foo-bar/"></button>

                # Without setting a value
                {% button %}{% button %}

                # Output:
                <button href="#"></button>
                
        **As parameter:**

        .. code-block:: python
        
                class Button(template.Component):
                    href = template.Attribute(default="#", name="data-href")

        Output:

        .. code-block::

                # base.html

                # Setting a value
                {% button href="foo-bar/" %}{% button %}

                # Output:
                <button data-href="foo-bar/"></button>

                # Without setting a value
                {% button %}{% button %}

                # Output:
                <button data-href="#"></button>

        """
        self.context_name = value

    def resolve(self, value: Union[Variable, FilterExpression, str], context, raise_exception: bool = True):
        """
        Resolves the template variable/expression specified as an argument, then checks if the value can be used.
        """
        try:
            return self.check_value(value.resolve(context), raise_exception=raise_exception)
        except VariableDoesNotExist as ex:
            raise self.VariableDoesNotExist(ex)



from . import Library
from .components import Slot

"""
Builtin tags 

* Slot: can be used inside any other component, therefore it can be pre-loaded inside django.
"""
register = Library()

register.tag('slot', Slot)



from django.template.exceptions import TemplateSyntaxError
from django.template.base import kwarg_re, FilterExpression, token_kwargs

from .components import Slot


def parse_component(nodelist, token, parser):
    """
    Load a component template and render it with the current context. You can pass
    additional context using keyword arguments.

    Example

    .. code-block::

        {% component %}{% endcomponent %}
        {% component with bar="BAZZ!" baz="BING!" %}{% endcomponent %}

    """
    bits = token.split_contents()

    # if len(bits) < 2:
    #     raise TemplateSyntaxError(
    #         "%r tag takes at least one argument: the name of the template to "
    #         "be included." % bits[0]
    #     )

    tag_name = bits.pop(0)
    args = []
    kwargs, options, slots = {}, {}, {}
    isolated_context = True

    while bits:
        bit = bits.pop(0)

        # if bit in options:
        #     raise TemplateSyntaxError('The %r option was specified more than once.' % bit)

        match = kwarg_re.match(bit)
        kwarg_format = match and match.group(1)

        if kwarg_format:
            key, value = match.groups()
            filter_expr = FilterExpression(value, parser)
            kwargs[key] = filter_expr
        else:
            filter_expr = FilterExpression(bit, parser)
            args.append(filter_expr)

        if bit == 'with':
            options = token_kwargs(bits, parser, support_legacy=False)
            if not options:
                raise TemplateSyntaxError('"with" in %r tag needs at least '
                                          'one keyword argument.' % tag_name)

    slot_nodes = list(filter(lambda x: isinstance(x[1], Slot), enumerate(nodelist)))

    while slot_nodes:
        pos, node = slot_nodes.pop()
        name = getattr(node, 'slot_name', pos)
        key = f'slot_{pos if name is None else name}'
        slots[key] = node
        del nodelist[pos]

    return tag_name, args, kwargs, options, slots, isolated_context



from typing import Optional, Union

from django.template import Context, RequestContext
from django.template.base import NodeList

from .helpers import format_attributes, format_value

__all__ = [
    'BaseContext',
    'Context',
    'TagContext',
    'ComponentContext'
]


class BaseContext:
    """
    Context wrapper which includes a couple of methods to create the same outcome.

    Attributes
    ----------
    initial: Union[Context, RequestContext, None]
        current template context
    dict_: dict
        extra values added to the context
    isolated: bool
        ensures that the context is isolated from the global context
    """

    default_class = Context

    def __init__(self, initial: Union[Context, RequestContext, None], dict_: dict, isolated: bool = True):
        if initial is None:
            self._wrap = self.default_class(dict_)
        elif isinstance(initial, (Context, RequestContext)):
            self._wrap = initial
        else:
            raise Exception('Cannot define initial')

        if isolated:
            self._wrap = self.new(dict_)
        else:
            self.update(dict_)

    def __copy__(self):
        return self._wrap.__copy__()

    def __repr__(self):
        return self._wrap.__repr__()

    def __iter__(self):
        return self._wrap.__iter__()

    def __setitem__(self, key, value):
        self._wrap.__setitem__(key, self.resolve(value))

    def __getitem__(self, key):
        """
        Get a variable's value, starting at the current context and going upward
        """
        return self._wrap.__getitem__(key)

    def __delitem__(self, key):
        """
        Delete a variable from the current context
        """
        self._wrap.__delitem__(key)

    def __contains__(self, key):
        return self._wrap.__contains__(key)

    def __eq__(self, other):
        """
        Compare two contexts by comparing theirs 'dicts' attributes.
        """
        return self._wrap.__eq__(other)

    def clean(self):
        """
        Return the wrapped context
        """
        return self._wrap

    def flatten(self) -> dict:
        """
        Return self.dicts as one dictionary.
        """
        return self.make().flatten()

    def push(self, *args, **kwargs):
        return self._wrap.push(*args, **kwargs)

    def pop(self):
        return self._wrap.pop()

    def set_upward(self, key, value):
        """
        Set a variable in one of the higher contexts if it exists there,
        otherwise in the current context.
        """
        self._wrap.set_upward(key, self.resolve(value))

    def get(self, key, otherwise=None):
        return self._wrap.get(key, otherwise)

    def setdefault(self, key, default=None):
        return self._wrap.setdefault(key, default)

    def new(self, values=None):
        """
        Return a new context with the same properties, but with only the
        values given in 'values' stored.
        """
        return self._wrap.new(values)

    def update(self, *args, **kwargs):
        return self._wrap.update(*args, **kwargs)

    def make(self):
        """
        Used to exec any function before compile the component tag.
        """
        return self.clean()

    def resolve(self, value):
        """
        Resolve the Variable/FilterExpression using the wrapped context
        """
        return format_value(value, self._wrap)


class TagContext(BaseContext):
    """
    Context used inside a tag function, if there is no need to create a component tag this is the way to
    recreate the same context.

    Attributes
    ----------
    attributes: Optional[dict]
        html attributes stored inside the context, and can be used as: {{ attributes }}
    initial: Optional[RequestContext]
        current template context
    isolated: bool
        ensures that the context is isolated from the global context
    **kwargs: dict
        extra values added to the context
    """

    default_class = RequestContext

    def __init__(self, attributes: Optional[dict] = None, initial: Optional[RequestContext] = None,
                 isolated: bool = False, **kwargs):
        super().__init__(initial, kwargs, isolated=isolated)
        self._attributes = {} if attributes is None else attributes

    @property
    def _attributes(self):
        return self['attributes']

    @_attributes.setter
    def _attributes(self, values: dict):
        if not isinstance(values, dict):
            raise Exception('Set attrs as a dict')
        self['attributes'] = values

    def add_class(self, *value):
        """
        Add a html "class" attribute inside the context.
        There can be used multiple times to store multiple classes in different scenarios.

        TODO: class as list to perform:
            - context.class.append
            - context.class = <var:list> | <var:string>
            - context.class.remove(<index:string>)
            - context.class.reset() === list()
        """

        values = [self.resolve(v) for v in value]

        try:
            self._attributes['class'].extend(values)
        except AttributeError:  # cannot extend
            classes = self._attributes['class']
            self._attributes['class'] = [classes] + values
        except KeyError:  # class not defined
            self._attributes['class'] = values

    def add_attribute(self, name: str, value):
        """
        Add a html attribute inside the context.
        There can be used multiple times to store multiple key/values in different scenarios.

        TODO: attribute as dict to perform:
          - context.attributes[<index:string>] = <value:string>
          - context.attributes = <var:dict>
          - del context.attributes[<index:string>]
          - context.attributes.reset() === dict()
        """
        value = self.resolve(value)

        # is this a class attribute?
        if name in 'class':
            return self.add_class(value)

        self._attributes[name] = value

    def make(self):
        """
        Collection of actions executed before render the component:
        - Format attributes as strings
        """
        context = super().make()
        context['attributes'] = format_attributes(self._attributes, context)
        return context


class ComponentContext(TagContext):
    """
    Context used inside component tags.

    Attributes
    ----------
    nodelist: NodeList
        nodelist passed to the component context
    initial: Optional[RequestContext]
        current template context
    attributes: Optional[dict]
        html attributes stored inside the context, and can be used as: {{ attributes }}
    isolated: bool
        ensures that the context is isolated from the global context
    **kwargs: dict
        extra values added to the context
    """
    def __init__(self, nodelist: NodeList, initial: RequestContext, attributes: Optional[dict] = None,
                 isolated: bool = True, **kwargs):
        super().__init__(attributes, initial=initial, isolated=isolated, **kwargs)
        self._nodelist = nodelist

        # TODO: get rid of this implementation
        # Make sure that request is part of the context
        if 'request' not in self and getattr(self._wrap, 'request', False):
            self['request'] = self._wrap.request

    def make(self):
        """
        Collection of actions executed before render the component:
        - Format attributes as strings
        - Render the current nodelist and store it inside the current context
        """
        context = super().make()
        context['nodelist'] = self._nodelist.render(context)
        return context



from django import template
from django.forms.widgets import Media
from django.template import Context

"""
Copyright (c) 2015 Jérôme Bon

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial 
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
DEALINGS IN THE SOFTWARE.

For more information about this: 

https://gitlab.com/Mojeer/django_components

"""


MEDIA_CONTEXT_KEY = "__django_component__media"


def media_tag(media_type):
    def media(parser, token):
        nodelist = parser.parse()
        return MediaNode(media_type, nodelist)

    return media


def ensure_media_context(root_context: dict):
    if MEDIA_CONTEXT_KEY not in root_context:
        root_context[MEDIA_CONTEXT_KEY] = Media()


def add_media(context: Context, media):
    root_context = context.dicts[0]
    ensure_media_context(root_context)
    root_context[MEDIA_CONTEXT_KEY] += media


class MediaNode(template.Node):
    def __init__(self, media_type, nodelist):
        self.media_type = media_type
        self.nodelist = nodelist

    def render(self, context):
        rendered = self.nodelist.render(context)
        return self.render_media(context) + rendered

    def render_media(self, context):
        tags = []
        if MEDIA_CONTEXT_KEY in context:
            media = context[MEDIA_CONTEXT_KEY]
            if self.media_type == "css":
                tags = media.render_css()
            elif self.media_type == "js":
                tags = media.render_js()
        return "".join(tags)
